@@ -7,7 +7,7 @@
 import unittest
 
 from xpra.util.objects import AdHocStruct
-from xpra.net.ssh.paramiko_client import keymd5
+from xpra.net.ssh.paramiko.util import keymd5
 from xpra.net.ssh.util import get_default_keyfiles
 
 
