@@ -4,7 +4,7 @@ import sys
 import gi
 gi.require_version('Gtk', '3.0')  # @UndefinedVariable
 gi.require_version('Gdk', '3.0')  # @UndefinedVariable
-from gi.repository import Gtk    #pylint: disable=wrong-import-position @UnresolvedImport
+from gi.repository import Gtk     # pylint: disable=wrong-import-position @UnresolvedImport
 
 width = 400
 height = 200
@@ -18,7 +18,7 @@ def main():
     window.connect("delete_event", Gtk.main_quit)
     window.realize()
     window.move(x, y)
-    from xpra.x11.gtk3.gdk_display_source import init_gdk_display_source
+    from xpra.x11.gtk.display_source import init_gdk_display_source
     init_gdk_display_source()
     from xpra.x11.bindings.window import X11WindowBindings  # @UnresolvedImport
     hints = {"position" : (x, y)}
