@@ -15,8 +15,9 @@ from PyQt6.QtNetwork import QTcpSocket
 from PyQt6.QtWidgets import QApplication
 
 from xpra import __version__
+from xpra.common import BACKWARDS_COMPATIBLE
 from xpra.exit_codes import ExitCode, ExitValue
-from xpra.net.common import PacketType
+from xpra.net.common import Packet
 from xpra.net.rencodeplus.rencodeplus import dumps, loads
 from xpra.net.protocol.header import pack_header, unpack_header, FLAGS_RENCODEPLUS
 
@@ -80,10 +81,12 @@ class Qt6Client:
             "session-id": uuid.uuid4().hex,
             "windows": True,
             "keyboard": True,
-            "mouse": True,
+            "pointer": True,
             "encodings": ("rgb32", "rgb24", "png", "jpg", "webp"),
             "network-state": False,  # tell older server that we don't have "ping"
         }
+        if BACKWARDS_COMPATIBLE:
+            hello["mouse"] = True
         self.send("hello", hello)
 
     def send(self, packet_type: str, *args) -> None:
@@ -123,62 +126,75 @@ class Qt6Client:
                 for index, data in self.raw_packets.items():
                     packet[index] = data
                 self.raw_packets = {}
-            self.process_packet(packet)
+            self.process_packet(Packet(*packet))
 
-    def process_packet(self, packet: PacketType) -> None:
-        packet_type_fn_name = str(packet[0]).replace("-", "_")
+    def process_packet(self, packet: Packet) -> None:
+        packet_type = packet.get_type()
+        packet_type_fn_name = packet_type.replace("-", "_")
         meth = getattr(self, f"_process_{packet_type_fn_name}", None)
         if not meth:
-            netlog.warn(f"Warning: missing handler for {packet[0]!r}")
+            netlog.warn(f"Warning: missing handler for {packet_type!r}")
             netlog("packet=%r", packet)
             return
         meth(packet)
 
-    def _process_hello(self, packet: PacketType) -> None:
-        self.hello = packet[1]
+    def _process_hello(self, packet: Packet) -> None:
+        self.hello = packet.get_dict(1)
         netlog.info("got hello from %s server" % self.hello.get("version", ""))
 
-    def _process_setting_change(self, packet: PacketType) -> None:
-        setting = packet[1]
+    def _process_setting_change(self, packet: Packet) -> None:
+        setting = packet.get_str(1)
         netlog.info(f"ignoring setting-change for {setting!r}")
 
-    def _process_startup_complete(self, _packet: PacketType) -> None:
+    def _process_startup_complete(self, _packet: Packet) -> None:
         netlog.info("client is connected")
 
-    def _process_encodings(self, packet: PacketType) -> None:
-        log(f"server encodings: {packet[1]}")
+    def _process_encodings(self, packet: Packet) -> None:
+        log(f"server encodings: {packet.get_strs(1)}")
 
-    def _process_new_window(self, packet: PacketType) -> None:
+    def _process_new_window(self, packet: Packet) -> None:
         self.new_window(packet)
 
-    def _process_new_override_redirect(self, packet: PacketType) -> None:
+    def _process_new_override_redirect(self, packet: Packet) -> None:
         self.new_window(packet, True)
 
-    def new_window(self, packet: PacketType, is_or=False) -> None:
+    def new_window(self, packet: Packet, is_or=False) -> None:
         from xpra.client.qt6.window import ClientWindow
-        wid, x, y, w, h = (int(item) for item in packet[1:6])
-        metadata = packet[6]
+        wid = packet.get_wid()
+        x = packet.get_i16(2)
+        y = packet.get_i16(3)
+        w = packet.get_i16(4)
+        h = packet.get_i16(5)
+        metadata = packet.get_dict(6)
         if is_or:
             metadata["override-redirect"] = is_or
         window = ClientWindow(self, wid, x, y, w, h, metadata)
         self.windows[wid] = window
         window.show()
 
-    def _process_lost_window(self, packet: PacketType) -> None:
-        wid = int(packet[1])
+    def _process_lost_window(self, packet: Packet) -> None:
+        wid = packet.get_wid()
         window = self.windows.get(wid)
         if window:
             window.close()
             del self.windows[wid]
 
-    def _process_raise_window(self, packet: PacketType) -> None:
-        wid = int(packet[1])
+    def _process_raise_window(self, packet: Packet) -> None:
+        wid = packet.get_wid()
         window = self.windows.get(wid)
         if window:
             window.raise_()
 
-    def _process_draw(self, packet: PacketType) -> None:
-        wid, x, y, width, height, coding, data, packet_sequence, rowstride = packet[1:10]
+    def _process_draw(self, packet: Packet) -> None:
+        wid = packet.get_wid()
+        x = packet.get_i16(2)
+        y = packet.get_i16(3)
+        width = packet.get_i16(4)
+        height = packet.get_i16(5)
+        coding = packet.get_str(6)
+        data = packet.get_buffer(7)
+        packet_sequence = packet.get_u64(8)
+        rowstride = packet.get_u32(9)
         window = self.windows.get(wid)
         now = monotonic()
         if window:
@@ -186,15 +202,15 @@ class Qt6Client:
             window.draw(x, y, width, height, coding, data, rowstride)
             decode_time = int(1000 * (monotonic() - now))
         else:
-            message = f"Warning: window {wid} not found"
+            message = f"Warning: window {wid:#x} not found"
             log.warn(message)
             decode_time = -1
         self.send("damage-sequence", packet_sequence, wid, width, height, decode_time, message)
 
-    def _process_window_metadata(self, packet: PacketType) -> None:
-        wid = packet[1]
-        metadata = packet[2]
-        log.info(f"window {wid}: {metadata}")
+    def _process_window_metadata(self, packet: Packet) -> None:
+        wid = packet.get_wid()
+        metadata = packet.get_dict(2)
+        log.info(f"window {wid:#x}: {metadata}")
 
     def update_focus(self, wid=0) -> None:
         if self.focused == wid:
@@ -242,7 +258,7 @@ class XpraQt6Client(Qt6Client):
         if not isinstance(conn, SocketConnection) or conn.socktype != "tcp":
             raise ValueError("only tcp socket connections are supported!")
 
-        def connect():
+        def connect() -> None:
             # self.socket.setSocketDescriptor(conn._socket.fileno())
             self.connect(*conn.remote)
 
