@@ -6,34 +6,32 @@
 # later version. See the file COPYING for details.
 
 import hashlib
+from collections.abc import Sequence
 from typing import Any
 
 from xpra.util.objects import typedict
 from xpra.util.str_fn import csv, Ellipsizer
 from xpra.util.env import envbool
-from xpra.os_util import gi_import
-from xpra.gtk.keymap import get_gtk_keymap, get_default_keymap
-from xpra.gtk.error import xsync, xlog
+from xpra.x11.error import xsync, xlog
 from xpra.keyboard.mask import (
     DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names,
 )
 from xpra.server.keyboard_config_base import KeyboardConfigBase
-from xpra.x11.gtk.keys import grok_modifier_map
 from xpra.x11.xkbhelper import (
     do_set_keymap, set_all_keycodes, set_keycode_translation,
     get_modifiers_from_meanings, get_modifiers_from_keycodes,
     clear_modifiers, set_modifiers, map_missing_modifiers,
-    clean_keyboard_state, get_keycode_mappings,
-    DEBUG_KEYSYMS,
+    clean_keyboard_state, get_keycode_mappings, get_keyval_mappings,
+    DEBUG_KEYSYMS, grok_modifier_map,
 )
 from xpra.x11.bindings.keyboard import X11KeyboardBindings
+from xpra.x11.bindings.test import XTestBindings
 from xpra.log import Logger
 
 log = Logger("keyboard")
 verboselog = Logger("keyboard", "verbose")
 
-Gdk = gi_import("Gdk")
-
+XTest = XTestBindings()
 X11Keyboard = X11KeyboardBindings()
 
 MAP_MISSING_MODIFIERS: bool = envbool("XPRA_MAP_MISSING_MODIFIERS", True)
@@ -90,7 +88,7 @@ class KeyboardConfig(KeyboardConfigBase):
         self.mod_pointermissing: list[str] = []
         self.mod_nuisance: set[str] = set(DEFAULT_MODIFIER_NUISANCE)
         self.keycodes = ()
-        self.x11_keycodes = []
+        self.x11_keycodes = {}
         self.layout = ""
         self.variant = ""
         self.options = ""
@@ -103,8 +101,11 @@ class KeyboardConfig(KeyboardConfigBase):
         self.keycode_translation = {}
         self.keycodes_for_modifier_keynames: dict[str, list[int]] = {}
         self.modifier_client_keycodes: dict[str, list] = {}
+        self.keycode_mappings: dict[int, list[str]] = {}
+        self.keyval_mappings = {}
+
+        self.compute_modifiers(True)
         self.compute_modifier_map()
-        self.keycode_mappings = {}
 
     def __repr__(self):
         return "KeyboardConfig(%s / %s / %s)" % (self.layout, self.variant, self.options)
@@ -225,7 +226,7 @@ class KeyboardConfig(KeyboardConfigBase):
                 hashadd(self.query_struct.get(k))
         return "%s/%s/%s/%s" % (self.layout, self.variant, self.options, m.hexdigest())
 
-    def compute_modifiers(self) -> None:
+    def compute_modifiers(self, initial=False) -> None:
         log("compute_modifiers() raw=%s", self.raw)
         if self.raw:
             with xsync:
@@ -256,7 +257,8 @@ class KeyboardConfig(KeyboardConfigBase):
                 if MAP_MISSING_MODIFIERS:
                     map_missing_modifiers(self.keynames_for_mod)
             else:
-                log.info("client did not supply any modifier definitions, using defaults")
+                if not initial:
+                    log.info("client did not supply any modifier definitions, using defaults")
                 self.keynames_for_mod = get_modifiers_from_meanings(DEFAULT_MODIFIER_MEANINGS)
                 if MAP_MISSING_MODIFIERS:
                     map_missing_modifiers(self.keynames_for_mod)
@@ -267,25 +269,23 @@ class KeyboardConfig(KeyboardConfigBase):
     def compute_modifier_keynames(self) -> None:
         self.keycodes_for_modifier_keynames = {}
         self.mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
-        keymap = get_default_keymap()
         if self.keynames_for_mod:
+            with xsync:
+                keysym_mappings = X11Keyboard.get_keysym_mappings()
             for modifier, keynames in self.keynames_for_mod.items():
                 for keyname in keynames:
                     if keyname in DEFAULT_MODIFIER_NUISANCE_KEYNAMES:
                         self.mod_nuisance.add(modifier)
-                    keyval = Gdk.keyval_from_name(keyname)
+                    keyval = X11Keyboard.parse_keysym(keyname)
+                    log("keyval(%s)=%s", keyname, keyval)
                     if keyval == 0:
-                        log.error("Error: no keyval found for keyname '%s' (modifier '%s')", keyname, modifier)
+                        log.error("Error: no keyval found for keyname %r (modifier %r)", keyname, modifier)
                         continue
-                    entries = keymap.get_entries_for_keyval(keyval)
-                    if entries:
-                        keycodes = []
-                        if entries[0] is True:
-                            keycodes = [entry.keycode for entry in entries[1]]
-                        for keycode in keycodes:
-                            l = self.keycodes_for_modifier_keynames.setdefault(keyname, [])
-                            if keycode not in l:
-                                l.append(keycode)
+                    keycodes = keysym_mappings.get(keyval, ())
+                    for keycode in keycodes:
+                        l = self.keycodes_for_modifier_keynames.setdefault(keyname, [])
+                        if keycode not in l:
+                            l.append(keycode)
         log("compute_modifier_keynames: keycodes_for_modifier_keynames=%s",
             self.keycodes_for_modifier_keynames)
 
@@ -356,7 +356,7 @@ class KeyboardConfig(KeyboardConfigBase):
             return
         log("set_keymap(%s) layout=%r, variant=%r, options=%r, query-struct=%r",
             translate_only, self.layout, self.variant, self.options, self.query_struct)
-        if translate_only:
+        if translate_only or not self.keycodes:
             self.keycode_translation = set_keycode_translation(self.x11_keycodes, self.keycodes)
             self.add_gtk_keynames()
             self.add_loose_matches()
@@ -364,6 +364,7 @@ class KeyboardConfig(KeyboardConfigBase):
             self.compute_modifier_keynames()
             self.compute_client_modifier_keycodes()
             self.update_keycode_mappings()
+            self.update_keyval_mappings()
             return
 
         with xlog:
@@ -413,13 +414,19 @@ class KeyboardConfig(KeyboardConfigBase):
             log("keyname_for_mod=%s", self.keynames_for_mod)
             clean_keyboard_state()
             self.update_keycode_mappings()
+            self.update_keyval_mappings()
             self.add_loose_matches()
 
     def add_gtk_keynames(self) -> None:
         # add the keynames we find via gtk
         # since we may rely on finding those keynames from the client
         # (used with non-native keymaps)
-        keymap = get_gtk_keymap()
+        try:
+            from xpra.gtk.keymap import get_gtk_keymap
+            keymap = get_gtk_keymap()
+        except ImportError:
+            log("add_gtk_keynames()", exc_info=True)
+            return
         log("add_gtk_keynames() gtk keymap=%s", Ellipsizer(keymap))
         for _, keyname, keycode, _, _ in keymap:
             if keyname not in self.keycode_translation:
@@ -473,18 +480,22 @@ class KeyboardConfig(KeyboardConfigBase):
                 for m in mappings:  # ie: (37, 'Control_L'), (105, 'Control_R')
                     if len(m) == 2:
                         keynames.append(m[1])  # ie: 'Control_L'
-                self.keynames_for_mod[modifier] = set(keynames)
+                self.keynames_for_mod[modifier] = keynames
             self.compute_modifier_keynames()
             self.compute_client_modifier_keycodes()
             log("set_default_keymap: keynames_for_mod=%s", self.keynames_for_mod)
             log("set_default_keymap: keycodes_for_modifier_keynames=%s", self.keycodes_for_modifier_keynames)
             log("set_default_keymap: modifier_map=%s", self.modifier_map)
             self.update_keycode_mappings()
+            self.update_keyval_mappings()
 
     def update_keycode_mappings(self) -> None:
         self.keycode_mappings = get_keycode_mappings()
 
-    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool, modifiers, keyval: int,
+    def update_keyval_mappings(self) -> None:
+        self.keyval_mappings = get_keyval_mappings()
+
+    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool, modifiers: list[str], keyval: int,
                        keystr: str, group: int) -> tuple[int, int]:
         if not self.enabled:
             kmlog(keyname, "ignoring keycode since keyboard is turned off")
@@ -500,7 +511,7 @@ class KeyboardConfig(KeyboardConfigBase):
         return self.find_matching_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)
 
     def find_matching_keycode(self, client_keycode: int, keyname: str,
-                              pressed: bool, modifiers, keyval: int, keystr: str, group: int) -> tuple[int, int]:
+                              pressed: bool, modifiers: list[str], keyval: int, keystr: str, group: int) -> tuple[int, int]:
         """
         from man xmodmap:
         The list of keysyms is assigned to the indicated keycode (which may be specified in decimal,
@@ -540,15 +551,30 @@ class KeyboardConfig(KeyboardConfigBase):
                 if name in ("ISO_Level3_Shift", "Mode_switch"):
                     mode = 1
                     break
-        levels = get_levels(mode, shift, group)
+
+        def toggle_modifier(mod: str) -> None:
+            keynames = self.keynames_for_mod.get(mod, set())
+            if keyname in keynames:
+                kml("not toggling '%s' since '%s' should deal with it", mod, keyname)
+                # the keycode we're returning is for this modifier,
+                # assume that this will end up doing what is needed
+                return
+            if mod in modifiers:
+                kml(f"removing {mod!r} from modifiers")
+                modifiers.remove(mod)
+            else:
+                kml(f"adding {mod} to modifiers")
+                modifiers.append(mod)
+
+        levels = get_levels(bool(mode), shift, bool(group))
+        level0 = levels[0]
         kml("will try levels: %s", levels)
         for level in levels:
             keycode = self.keycode_translation.get((keyname, level), -1)
             if keycode < 0:
                 continue
-            keysyms = self.keycode_mappings.get(keycode)
+            keysyms = self.keycode_mappings.get(keycode, ())
             klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
-            level0 = levels[0]
             uq_keysyms = set(keysyms)
             if len(uq_keysyms) <= 1 or (len(keysyms) > level0 and keysyms[level0] == ""):
                 # if the keysym we would match for this keycode is 'NoSymbol',
@@ -557,20 +583,6 @@ class KeyboardConfig(KeyboardConfigBase):
                 kml("not toggling any modifiers state for keysyms=%s", keysyms)
                 break
 
-            def toggle_modifier(mod: str) -> None:
-                keynames = self.keynames_for_mod.get(mod, set())
-                if keyname in keynames:
-                    kml("not toggling '%s' since '%s' should deal with it", mod, keyname)
-                    # the keycode we're returning is for this modifier,
-                    # assume that this will end up doing what is needed
-                    return
-                if mod in modifiers:
-                    kml(f"removing {mod!r} from modifiers")
-                    modifiers.remove(mod)
-                else:
-                    kml(f"adding {mod} to modifiers")
-                    modifiers.append(mod)
-
             # keypad overrules shift state (see #2702):
             if keyname.startswith("KP_"):
                 if numlock_modifier and not numlock:
@@ -593,21 +605,41 @@ class KeyboardConfig(KeyboardConfigBase):
         if keycode < 0:
             keycode = self.keycode_translation.get(keyname, -1)
             klog("=%i, %i (keyname translation)", keycode, rgroup)
-        if keycode < 0:
+            if keycode < 0:
+                # could this be a modifier?
+                modname = {"ctrl": "control"}.get(keyname.lower(), keyname.lower())
+                for modkey in self.keynames_for_mod.get(modname, ()):
+                    keycode = self.keycode_translation.get(modkey, -1)
+                    if keycode >= 0:
+                        break
+            if keycode < 0:
+                lowercase = dict((x.lower(), x) for x in self.keycode_translation.keys() if isinstance(x, str))
+                match = lowercase.get(keyname, "")
+                if match:
+                    keycode = self.keycode_translation.get(match, -1)
+        # noinspection PyChainedComparisons
+        if keycode < 0 and keyval > 0:
             # last resort, find using the keyval:
-            keymap = get_default_keymap()
-            b, entries = keymap.get_entries_for_keyval(keyval)
-            if b and entries:
-                for entry in entries:
-                    if entry.group == group:
-                        return entry.keycode, entry.group
+            group_mapping = self.keyval_mappings.get(keyval, {})
+            if group_mapping:
+                # this keyval was found!
+                # try to preserve the group:
+                keycodes = group_mapping.get(keyval, [])
+                if keycodes:
+                    keycode = keycodes[0]
+                else:
+                    # try other groups:
+                    for group, keycodes in group_mapping.items():
+                        if keycodes:
+                            keycode = keycodes[0]
+                            rgroup = group
         return keycode, rgroup
 
     def get_current_mask(self) -> list[str]:
         # this is only called from an existing xsync context in this module
         return mask_to_names(X11Keyboard.query_mask(), self.modifier_map)
 
-    def make_keymask_match(self, modifier_list, ignored_modifier_keycode=None, ignored_modifier_keynames=None) -> None:
+    def make_keymask_match(self, modifier_list, ignored_modifier_keycode=0, ignored_modifier_keynames=None) -> None:
         """
             Given a list of modifiers that should be set, try to press the right keys
             to make the server's modifier list match it.
@@ -633,22 +665,22 @@ class KeyboardConfig(KeyboardConfigBase):
             return
         if ignored_modifier_keynames is None:
             # this is not a keyboard event, ignore modifiers in "mod_pointermissing"
-            def is_ignored(modifier, _modifier_keynames) -> bool:
-                return modifier in (self.mod_pointermissing or [])
+            def is_ignored(modifier: str, _modifier_keynames: Sequence[str]) -> bool:
+                return modifier in (self.mod_pointermissing or ())
         else:
             # keyboard event: ignore the keynames specified
             # (usually the modifier key being pressed/unpressed)
-            def is_ignored(_modifier, modifier_keynames) -> bool:
-                return len(set(modifier_keynames or []) & set(ignored_modifier_keynames or [])) > 0
+            def is_ignored(_modifier: str, modifier_keynames: Sequence[str]) -> bool:
+                return len(set(modifier_keynames or ()) & set(ignored_modifier_keynames or ())) > 0
 
-        def filtered_modifiers_set(modifiers) -> set[str]:
+        def filtered_modifiers_set(modifiers: Sequence[str]) -> set[str]:
             m: set[str] = set()
             mm = self.mod_managed or ()
             for modifier in modifiers:
                 if modifier in mm:
                     log("modifier is server managed: %s", modifier)
                     continue
-                keynames = self.keynames_for_mod.get(modifier)
+                keynames = self.keynames_for_mod.get(modifier, ())
                 if is_ignored(modifier, keynames):
                     log("modifier '%s' ignored (in ignored keynames=%s)", modifier, keynames)
                     continue
@@ -656,79 +688,6 @@ class KeyboardConfig(KeyboardConfigBase):
             verboselog("filtered_modifiers_set(%s)=%s", modifiers, m)
             return m
 
-        def change_mask(modifiers: set[str], press: bool, info: str) -> list[str]:
-            failed = []
-            for modifier in modifiers:
-                if modifier not in self.keynames_for_mod:
-                    log.error(f"Error: unknown modifier {modifier!r}")
-                    log.error(" known modifiers: %s", csv(self.keynames_for_mod.keys()))
-                    continue
-                keynames = self.keynames_for_mod.get(modifier, ())
-                if not keynames:
-                    log.warn("Warning: found no keynames for '%s'", modifier)
-                    continue
-                # find the keycodes that match the keynames for this modifier
-                keycodes = []
-                # log.info("keynames(%s)=%s", modifier, keynames)
-                for keyname in keynames:
-                    if keyname in self.keys_pressed.values():
-                        # found the key which was pressed to set this modifier
-                        for keycode, name in self.keys_pressed.items():
-                            if name == keyname:
-                                log("found the key pressed for %s: %s", modifier, name)
-                                keycodes.insert(0, keycode)
-                    keycodes_for_keyname = self.keycodes_for_modifier_keynames.get(keyname, [])
-                    for keycode in keycodes_for_keyname:
-                        if keycode not in keycodes:
-                            keycodes.append(keycode)
-                if ignored_modifier_keycode is not None and ignored_modifier_keycode in keycodes:
-                    log("modifier '%s' ignored (ignored keycode=%s)", modifier, ignored_modifier_keycode)
-                    continue
-                # nuisance keys (lock, num, scroll) are toggled by a
-                # full key press + key release (so act accordingly in the loop below)
-                nuisance = modifier in self.mod_nuisance
-                log("keynames(%s)=%s, keycodes=%s, nuisance=%s, nuisance keys=%s",
-                    modifier, keynames, keycodes, nuisance, self.mod_nuisance)
-                modkeycode = None
-                if not press:
-                    # since we want to unpress something,
-                    # let's try the keycodes we know are pressed first:
-                    kdown = X11Keyboard.get_keycodes_down()
-                    pressed = [x for x in keycodes if x in kdown]
-                    others = [x for x in keycodes if x not in kdown]
-                    keycodes = pressed + others
-                for keycode in keycodes:
-                    if nuisance:
-                        X11Keyboard.xtest_fake_key(keycode, True)
-                        X11Keyboard.xtest_fake_key(keycode, False)
-                    else:
-                        X11Keyboard.xtest_fake_key(keycode, press)
-                    if VERIFY_MODIFIERS:
-                        new_mask = self.get_current_mask()
-                        success = (modifier in new_mask) == press
-                    else:
-                        success = True
-                    if success:
-                        modkeycode = keycode
-                        log("change_mask(%s) %s modifier '%s' using keycode %s",
-                            info, modifier_list, modifier, keycode)
-                        break  # we're done for this modifier
-                    log("%s %s with keycode %s did not work", info, modifier, keycode)
-                    if press and not nuisance:
-                        log(" trying to unpress it!")
-                        X11Keyboard.xtest_fake_key(keycode, False)
-                        # maybe doing the full keypress (down+up) worked:
-                        new_mask = self.get_current_mask()
-                        if (modifier in new_mask) == press:
-                            modkeycode = keycode
-                            break
-                    log("change_mask(%s) %s modifier '%s' using keycode %s, success: %s",
-                        info, modifier_list, modifier, keycode, success)
-                if not modkeycode:
-                    failed.append(modifier)
-            log("change_mask(%s, %s, %s) failed=%s", modifiers, press, info, failed)
-            return failed
-
         with xsync:
             current = filtered_modifiers_set(self.get_current_mask())
             wanted = filtered_modifiers_set(modifier_list or [])
@@ -737,8 +696,8 @@ class KeyboardConfig(KeyboardConfigBase):
             log("make_keymask_match(%s) current mask: %s, wanted: %s, ignoring=%s/%s, keys_pressed=%s",
                 modifier_list, current, wanted,
                 ignored_modifier_keycode, ignored_modifier_keynames, self.keys_pressed)
-            fr = change_mask(current.difference(wanted), False, "remove")
-            fa = change_mask(wanted.difference(current), True, "add")
+            fr = self.change_mask(current.difference(wanted), False, "remove", ignored_modifier_keycode)
+            fa = self.change_mask(wanted.difference(current), True, "add", ignored_modifier_keycode)
             if not fr or fa:
                 return
             if fr:
@@ -749,10 +708,79 @@ class KeyboardConfig(KeyboardConfigBase):
                 log.warn(" %s", csv(fa))
             # this should never happen.. but if it does?
             # something didn't work, use the big hammer and start again from scratch:
-            log.warn(" keys still pressed=%s", X11Keyboard.get_keycodes_down())
-            X11Keyboard.unpress_all_keys()
+            keycodes = X11Keyboard.get_keycodes_down()
+            log.warn(" keys still pressed=%s", keycodes)
+            XTest.unpress_keys(tuple(keycodes.keys()))
             log.warn(" doing a full keyboard reset, keys now pressed=%s", X11Keyboard.get_keycodes_down())
             # and try to set the modifiers one last time:
             current = filtered_modifiers_set(self.get_current_mask())
-            change_mask(current.difference(wanted), False, "remove")
-            change_mask(wanted.difference(current), True, "add")
+            self.change_mask(current.difference(wanted), False, "remove", ignored_modifier_keycode)
+            self.change_mask(wanted.difference(current), True, "add", ignored_modifier_keycode)
+
+    def change_mask(self, modifiers: set[str], press: bool, info: str, ignored_modifier_keycode=0) -> list[str]:
+        failed: list[str] = []
+        for modifier in modifiers:
+            if modifier not in self.keynames_for_mod:
+                log.error(f"Error: unknown modifier {modifier!r}")
+                log.error(" known modifiers: %s", csv(self.keynames_for_mod.keys()))
+                continue
+            keynames = self.keynames_for_mod.get(modifier, ())
+            if not keynames:
+                log.warn("Warning: found no keynames for '%s'", modifier)
+                continue
+            # find the keycodes that match the keynames for this modifier
+            keycodes = []
+            for keyname in keynames:
+                # find the key which was pressed to set this modifier
+                for keycode, name in self.keys_pressed.items():
+                    if name == keyname:
+                        log("found the key pressed for %s: %s", modifier, name)
+                        keycodes.insert(0, keycode)
+                keycodes_for_keyname = self.keycodes_for_modifier_keynames.get(keyname, ())
+                for keycode in keycodes_for_keyname:
+                    if keycode not in keycodes:
+                        keycodes.append(keycode)
+            if ignored_modifier_keycode and ignored_modifier_keycode in keycodes:
+                log("modifier '%s' ignored (ignored keycode=%s)", modifier, ignored_modifier_keycode)
+                continue
+            # nuisance keys (lock, num, scroll) are toggled by a
+            # full key press + key release (so act accordingly in the loop below)
+            nuisance = modifier in self.mod_nuisance
+            log("keynames(%s)=%s, keycodes=%s, nuisance=%s, nuisance keys=%s",
+                modifier, keynames, keycodes, nuisance, self.mod_nuisance)
+            modkeycode = 0
+            if not press:
+                # since we want to unpress something,
+                # let's try the keycodes we know are pressed first:
+                kdown = X11Keyboard.get_keycodes_down()
+                pressed = [x for x in keycodes if x in kdown]
+                others = [x for x in keycodes if x not in kdown]
+                keycodes = pressed + others
+            for keycode in keycodes:
+                if nuisance:
+                    XTest.xtest_fake_key(keycode, True)
+                    XTest.xtest_fake_key(keycode, False)
+                else:
+                    XTest.xtest_fake_key(keycode, press)
+                new_press = modifier in self.get_current_mask()
+                success = not VERIFY_MODIFIERS or new_press == press
+                if success:
+                    modkeycode = keycode
+                    log("change_mask(%s) modifier '%s' using keycode %s",
+                        info, modifier, keycode)
+                    break  # we're done for this modifier
+                log("%s %s with keycode %s did not work", info, modifier, keycode)
+                if press and not nuisance:
+                    log(" trying to unpress it!")
+                    XTest.xtest_fake_key(keycode, False)
+                    # maybe doing the full keypress (down+up) worked:
+                    new_press = modifier in self.get_current_mask()
+                    if new_press == press:
+                        modkeycode = keycode
+                        break
+                log("change_mask(%s) modifier '%s' using keycode %s, success: %s",
+                    info, modifier, keycode, success)
+            if not modkeycode:
+                failed.append(modifier)
+        log("change_mask(%s, %s, %s) failed=%s", modifiers, press, info, failed)
+        return failed
